# ------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ------------------------------------------------------------------------------
import math
import datetime


class FormatWebHDFS():
    @staticmethod
    def _format_perm_octet_as_rwx(perm_octets):
        access_octet = \
        {
            '0': ('---'),
            '1': ('--x'),
            '2': ('-w-'),
            '3': ('-wx'),
            '4': ('r--'),
            '5': ('r-x'),
            '6': ('rw-'),
            '7': ('rwx')
        }

        result = ""
        for char in perm_octets:
            result += access_octet[char]

        return result

    @staticmethod
    def _format_time(time):
        # Note that HDFS timestamps are in milliseconds so the value needs to
        # be divided by 1000
        time = time / 1000
        result = datetime.datetime.utcfromtimestamp(int(time))

        # Note on newer python distributions the format strings below throw an
        # exception.
        # For now don't change the format and just swallow the error.
        #
        try:
            if result.year == datetime.date.today().year:
                result_str = result.strftime('%_b %d %H:%M')
            else:
                result_str = result.strftime('%_b %_d  %_Y')
        except:
            result_str = str(result)

        return result_str

    @staticmethod
    def _format_file_size_in_bytes(size_bytes):
        size_name = ("", "KB", "MB", "GB", "TB", "PB", "EB", "ZB")
        if size_bytes != 0:
            magnitude = int(math.floor(math.log(size_bytes, 1024)))
            power = math.pow(1024, magnitude)
            magnitude_size = round(size_bytes / power, 2)
            if magnitude == 0:
                file_size_string = "{0:>6}".format(int(magnitude_size))
            else:
                file_size_string = "{0:>4}{1:2}".format(magnitude_size,
                                                        size_name[magnitude])
        else:
            file_size_string = '0'
        return file_size_string

    @staticmethod
    def convert_filestatuses_to_list(filestatuses):
        table = []

        for filestatus in filestatuses['FileStatus']:
            row = {}
            row['permission'] = FormatWebHDFS._format_perm_octet_as_rwx(
                filestatus['permission'])
            row['owner'] = filestatus['owner']
            row['group'] = filestatus['group']
            row['size'] = FormatWebHDFS._format_file_size_in_bytes(
                filestatus['length'])
            row['datetime'] = FormatWebHDFS._format_time(
                filestatus['modificationTime'])
            if filestatus['type'] == 'DIRECTORY':
                row['name'] = filestatus['pathSuffix'] + '/'
            else:
                row['name'] = filestatus['pathSuffix']

            table.append(row)
        return table

    @staticmethod
    def convert_filestatuses_to_name_list(filestatuses):
        table = []

        for filestatus in filestatuses['FileStatus']:
            if filestatus['type'] == 'DIRECTORY':
                name = filestatus['pathSuffix'] + '/'
            else:
                name = filestatus['pathSuffix']

            table.append(name)
        return table

    @staticmethod
    def format_directory_as_output(directory):
        """
        Format a FileStatuses structure into a Linux ll command structure.

        The format is:
            Permissions Owner Group Size Date Time Name
        """
        result = ""
        if len(directory) > 0:
            owner_max_len = max([len(x['owner']) for x in directory])
            group_max_len = max([len(x['group']) for x in directory])
            format_string = "{{0}} {{1:>{0}}} {{2:>{1}}} {{3:>8}} {{4}} " \
                            "{{5}}\n".format(owner_max_len, group_max_len)

            for entry in directory:
                result += format_string.format(
                    entry['permission'],
                    entry['owner'],
                    entry['group'],
                    entry['size'],
                    entry['datetime'],
                    entry['name']
                )
        return result


def get_credentials_json(credentials_string):
    """
    Get the JSON representation of the credentials.
    :param credentials_string: string containing comma-separated key=value
           pairs representing the credentials. The first "="
    :return: the JSON representation of the credentials as a string.
    """
    from json import JSONEncoder
    import json
    import re

    if (credentials_string is None) or len(credentials_string) == 0:
        return None

    pairs = re.split(r'(?<!\\),', credentials_string)

    kvpairs = dict()

    for pair in pairs:
        # split on = ignored the escaped ones
        try:
            # remove escaped characters for ,
            pair = pair.replace('\\,', ',').strip()
            (key, value) = pair.split('=', 1)
            kvpairs[key] = value
        except:
            raise ValueError('Bad formatting of credentials at {}'.format(pair))
    credentials_dict = {'credentials': kvpairs}

    return json.loads(JSONEncoder().encode(credentials_dict))
