# ------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ------------------------------------------------------------------------------

from .api import DefaultApi
from azdata.cli.core.exceptions import (
    ApiError,
    CliError,
    ControllerError,
    http_status_codes
)
from datetime import datetime, timezone
import uuid
import json

__all__ = ['DuskyControllerService']

# ---------------------------------------------------------------------------- #
# Dusky router: /dusky/*
# ---------------------------------------------------------------------------- #

class DuskyControllerService(object):
    
    def postgres_server_backup_create(self, namespace, server_group_id, backup_name):
        """
        Creates Arc Postgres backup.
        :return:
        """
        try:
            api = DefaultApi(api_client=self.api_client)
            response = api.postgres_server_backup_create(namespace=namespace, server_group_id=server_group_id, name=backup_name)
            return self._format_response('create the backup', response)
        except ApiError as e:
            e.headers = None
            raise ControllerError(e, status=e.status)

    def postgres_server_backup_restore(self, namespace, server_group_id, backup_id):
        """
        Restores Arc Postgres backup.
        :return:
        """
        try:
            api = DefaultApi(api_client=self.api_client)
            response = api.postgres_server_backup_restore(namespace=namespace, server_group_id=server_group_id, backup_id=backup_id)
            return self._format_response('restore the backup', response)
        except ApiError as e:
            e.headers = None
            raise ControllerError(e, status=e.status)

    def postgres_server_backup_restore_status(self, namespace, server_group_id):
        """
        Restores Arc Postgres backup.
        :return:
        """
        try:
            api = DefaultApi(api_client=self.api_client)
            response = api.postgres_server_backup_restore_status(namespace=namespace, server_group_id=server_group_id)
            return self._format_response('get the restore status', response)
        except ApiError as e:
            e.headers = None
            raise ControllerError(e, status=e.status)
    
    def postgres_server_backup_show(self, namespace, server_group_id, backup_id):
        """
        Gets Arc Postgres backup.
        :return:
        """
        try:
            api = DefaultApi(api_client=self.api_client)
            response = api.postgres_server_backup_show(namespace=namespace, server_group_id=server_group_id, backup_id=backup_id)
            return self._format_response('get the backup', response)
        except ApiError as e:
            e.headers = None
            raise ControllerError(e, status=e.status)
    
    def postgres_server_backup_list(self, namespace, server_group_id):
        """
        Lists Arc Postgres backups.
        :return:
        """
        try:
            api = DefaultApi(api_client=self.api_client)
            backups = api.list_dusky_backups(ns=namespace, server_group_id=server_group_id, version="1", resource_type="dusky")
            return [self._format_response('list backups', b) for b in backups.get('backups', [])]
        except ApiError as e:
            e.headers = None
            raise ControllerError(e, status=e.status)

    def postgres_server_backup_delete(self, namespace, server_group_id, backup_id):
        """
        Delete Arc Postgres backup for the given backup id.
        :return:
        """
        try:
            api = DefaultApi(api_client=self.api_client)
            backup = api.delete_dusky_backup(ns=namespace, server_group_id=server_group_id, backup_id=backup_id, version="1", resource_type="dusky")
            return self._format_response('delete backup', backup)
        except ApiError as e:
            e.headers = None
            raise ControllerError(e, status=e.status, response_json=e.response_json)

    def _map_progress(self, state):
        """
        Maps the Grpc status code into a human-readable state
        :return:
        """
        if state == 0:
            return 'Pending'
        if state == 1:
            return 'Active'
        if state == 2:
            return 'Done'
        if state == 3:
            return 'Failed'
        else:
            return 'Unknown'
    
    def _format_response(self, msg, response):
        """
        Formats the returned response for easier client parsing
        :return:
        """
        status = response.get('status', {})
        if not status.get('result', True):
            message = status.get('message')
            if message is not None:
                raise CliError('Failed to {}. {}'.format(msg, message))
            else:
                status.pop('result', None)
                raise CliError('Failed to {}.\n{}'.format(msg, json.dumps(status, indent=2)))

        response["progress"] = self._map_progress(response.get("progress"))
        response.setdefault('receipt', {}).setdefault('id', '')
        response['receipt']['id'].replace('-', '')
        response['timestamp'] = str(datetime.fromtimestamp(response['timestamp']['seconds'], tz=timezone.utc))

        return response
