# ------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ------------------------------------------------------------------------------

from azdata.cli.core.exceptions import CliError
from azdata.cli.core.util import(
    FileUtil,
    is_windows)
from azdata.cli.core.exceptions import http_status_codes
from azdata.cli.core.clients.kubernetes_client import (
    K8sApiException)
from azdata.cli.core.models.data_controller_custom_resource import DataControllerCustomResource
from azdata.cli.core.models.custom_resource_definition import CustomResourceDefinition
from azdata.cli.core.constants import (
    AZDATA_PASSWORD,
    MGMT_PROXY,
    ARC_GROUP,
    ARC_API_VERSION,
    DATA_CONTROLLER_PLURAL)
from azdata.cli.core.prompt import (prompt, prompt_pass)
from azdata.cli.commands.postgres.constants import (
    RESOURCE_KIND_11,
    RESOURCE_KIND_12,
    RESOURCE_CATEGORY,
    COMMAND_UNIMPLEMENTED,
    API_VERSION,
    API_GROUP)
from .models.postgres_cr_model import PostgresqlCustomResource
from azdata.cli.core.exceptions import KubernetesError
from azdata.cli.core.deploy import DeploymentConfigUtil
from collections import OrderedDict
from humanfriendly.terminal.spinners import AutomaticSpinner
from kubernetes import client as k8sClient

import azdata.cli.core.deploy as util
import copy
import json
import time
import os
import base64
import yaml
import sys

from azdata.cli.core.logging import get_logger
from urllib3.exceptions import NewConnectionError, MaxRetryError

CONNECTION_RETRY_ATTEMPTS = 12
RETRY_INTERVAL = 5

logger = get_logger(__name__)


# ------------------------------------------------------------------------------
# Server Commands
# ------------------------------------------------------------------------------

def arc_postgres_server_create(client,
                               name,
                               path=None,
                               # namespace=None,
                               cores_limit=None,
                               cores_request=None,
                               memory_limit=None,
                               memory_request=None,
                               storage_class_data=None,
                               storage_class_logs=None,
                               storage_class_backups=None,
                               extensions=None,
                               volume_size_data=None,
                               volume_size_logs=None,
                               volume_size_backups=None,
                               workers=None,
                               engine_version=None,
                               no_external_endpoint=None,
                               dev=None,
                               port=None,
                               no_wait=False,
                               engine_settings=None):
    """
    Create a PostgreSQL server group.
    :param client:
    :param path: The src filepath of the postgres resource.
    :param name: The name of the instance.
    :param namespace: Namespace where the PostgreSQL server group is deployed.
    :param cores_limit: The limit of cores of the managed instance in integer number of vcores.
    :param cores_request: The request for cores of the managed instance in integer number of vcores.
    :param memory_limit: The limit of the capacity of the managed instance in integer amount of memory in GBs.
    :param memory_request: The request for the capacity of the managed instance in integer amount of memory in GBs.
    :param storage_class_data: The storage classes to be used for data persistent volumes.
    :param storage_class_logs: The storage classes to be used for logs persistent volumes.
    :param storage_class_backups: The storage classes to be used for backups persistent volumes.
    :param volume_size_data: The volume size for the storage classes to be used for data.
    :param volume_size_logs: The volume size for the storage classes to be used for logs.
    :param volume_size_backups: The volume size for the storage classes to be used for backups.
    :param extensions: A comma-separated list of Postgres extensions that should be enabled.
    :param workers: The number of worker nodes to provision in a sharded cluster, or zero for single-node Postgres.
    :param no_external_endpoint: If not specified, an external service is created using the same service type as the dc.
    :param dev: If this is specified, then it is considered a dev instance and will not be billed for.
    :param port: Optional parameter for the service port.
    :param no_wait: If given, the command won't wait until the deployment is ready before returning.
    :param engine_settings: If given, sets the engine properties
    :return:
    """
    try:
        util.check_and_set_kubectl_context()

        # TODO: Support user supplied namespace when the backend supports it
        namespace = client.profile.active_context.namespace

        if len(_get_postgres_custom_objects(client, name, namespace)) > 0:
            raise CliError("A PostgreSQL server group with the name {} already exists.".format(name))

        # If no template files are provided, retrieve them from the Arc Data Controller.
        if not path:
            resource_kind = _get_resource_kind_by_engine_version(engine_version)
            spec_object = util.retry(lambda: client.apis.controller.resource_spec_get(resource_kind),
                                     retry_count=CONNECTION_RETRY_ATTEMPTS,
                                     retry_delay=RETRY_INTERVAL,
                                     retry_method="resource spec get",
                                     retry_on_exceptions=(NewConnectionError, MaxRetryError))

            cr = PostgresqlCustomResource(spec_object)
        # Otherwise, use the provided src file.
        else:
            json_object = FileUtil.read_json(path)
            cr = PostgresqlCustomResource(json_object)

            if engine_version is None:
                resource_kind = cr.body["kind"]
            else:
                # Override resource kind in the spec.
                resource_kind = _get_resource_kind_by_engine_version(engine_version)
                cr.body["kind"] = resource_kind

        # Retrieve postgres CRD
        crd = CustomResourceDefinition(client.apis.controller.resource_crd_get(resource_kind))
        resource_kind_plural = crd.plural

        # Name required
        cr.name = name

        cr.namespace = namespace

        custom_object_exists = util.retry(lambda: client.apis.kubernetes.namespaced_custom_object_exists(
                                                    cr.name, cr.namespace,
                                                    group=API_GROUP,
                                                    version=API_VERSION,
                                                    plural=resource_kind_plural),
                                          retry_count=CONNECTION_RETRY_ATTEMPTS,
                                          retry_delay=RETRY_INTERVAL, retry_method="get namespaced custom object",
                                          retry_on_exceptions=(NewConnectionError, MaxRetryError, KubernetesError))

        if custom_object_exists:
            raise ValueError("Postgres Server `{}` already exists in namespace `{}`.".format(name, namespace))

        # add scheduling settings
        if cores_limit:
            cr.default_cores_limit = cores_limit

        if cores_request:
            cr.default_cores_request = cores_request

        if memory_limit:
            cr.default_memory_limit = memory_limit

        if memory_request:
            cr.default_memory_request = memory_request

        # Validate cores/memory after both have been set to avoid behavior related to the order of property setting
        #
        cr.validate_memory()
        cr.validate_cores()

        if storage_class_data:
            if not client.apis.kubernetes.storage_class_exists(storage_class_data):
                raise ValueError("Storage class \"%s\" does not exist" % storage_class_data)
            cr.storage_class_data = storage_class_data

        if storage_class_logs:
            if not client.apis.kubernetes.storage_class_exists(storage_class_logs):
                raise ValueError("Storage class \"%s\" does not exist" % storage_class_logs)
            cr.storage_class_logs = storage_class_logs

        if storage_class_backups:
            if not client.apis.kubernetes.storage_class_exists(storage_class_backups):
                raise ValueError("Storage class \"%s\" does not exist" % storage_class_backups)
            cr.storage_class_backups = storage_class_backups

        if volume_size_data:
            cr.volume_size_data = volume_size_data

        if volume_size_logs:
            cr.volume_size_logs = volume_size_logs

        if volume_size_backups:
            cr.volume_size_backups = volume_size_backups

        if workers is not None:
            cr.shards = workers

        if extensions:
            extensions = [e.strip() for e in extensions.split(',')]
        else:
            extensions = cr.extensions if extensions is None else []
        if cr.shards and 'citus' not in extensions: # automatically add the 'citus' plugin if sharding is enabled
            extensions.insert(0, 'citus')
        cr.extensions = extensions

        if not no_external_endpoint:
            response = util.retry(lambda: client.apis.kubernetes.list_namespaced_custom_object(
                                    namespace,
                                    group=ARC_GROUP,
                                    version=ARC_API_VERSION,
                                    plural=DATA_CONTROLLER_PLURAL),
                                  retry_count=CONNECTION_RETRY_ATTEMPTS,
                                  retry_delay=RETRY_INTERVAL,
                                  retry_method="list namespaced custom object",
                                  retry_on_exceptions=(NewConnectionError, MaxRetryError, K8sApiException))

            dcs = response.get("items")
            if not dcs:
                raise CliError("No data controller exists in namespace `{}`. Cannot set external endpoint argument."
                               .format(namespace))
            else:
                dc_cr = DataControllerCustomResource(dcs[0])
                cr.service_type = dc_cr.controller_service_type

        if port:
            cr.service_port = port

        if dev:
            cr.dev = dev

        if engine_settings is not None:
            engine_settings_tuple = (engine_settings, None)
            cr.default_engine_settings = engine_settings_tuple

        util.retry(lambda: client.apis.kubernetes.create_custom_resource_definition(crd),
                   retry_count=CONNECTION_RETRY_ATTEMPTS,
                   retry_delay=RETRY_INTERVAL, retry_method="create custom resource definition",
                   retry_on_exceptions=(NewConnectionError, MaxRetryError, K8sApiException, KubernetesError))

        secret_name = name + "-login-secret"

        secret_exists = util.retry(lambda: client.apis.kubernetes.secret_exists(cr.namespace, secret_name),
                                   retry_count=CONNECTION_RETRY_ATTEMPTS,
                                   retry_delay=RETRY_INTERVAL, retry_method="secret exists",
                                   retry_on_exceptions=(NewConnectionError, MaxRetryError, K8sApiException))

        if not secret_exists:
            pw = os.environ.get(AZDATA_PASSWORD)
            if not pw:
                if sys.stdin.isatty():
                    pw = prompt_pass("Postgres Server password:", confirm=True, allow_empty=False)
                else:
                    raise ValueError("Please provide a Postgres Server password through the env "
                                     "variable AZDATA_PASSWORD.")
            else:
                client.stdout("Using AZDATA_PASSWORD environment variable for `{}` password.".format(name))

            model = {"secretName": secret_name}
            encoding = "utf-8"
            model["base64Username"] = base64.b64encode(
                bytes("postgres", encoding)).decode(encoding)
            model["base64Password"] = base64.b64encode(
                bytes(pw, encoding)).decode(encoding)
            temp = util.get_config_from_template(
                os.path.join(os.path.dirname(os.path.realpath(__file__)), "templates", "postgres-login.yaml.tmpl"),
                model)
            postgres_secret = yaml.full_load(temp)

            util.retry(
                lambda: client.apis.kubernetes.create_secret(cr.namespace, postgres_secret, ignore_conflict=True),
                retry_count=CONNECTION_RETRY_ATTEMPTS,
                retry_delay=RETRY_INTERVAL, retry_method="create secret",
                retry_on_exceptions=(NewConnectionError, MaxRetryError, K8sApiException))

        util.retry(lambda: client.apis.kubernetes.create_namespaced_custom_object(crd, cr, ignore_conflict=True),
                   retry_count=CONNECTION_RETRY_ATTEMPTS,
                   retry_delay=RETRY_INTERVAL, retry_method="create namespaced custom object",
                   retry_on_exceptions=(NewConnectionError, MaxRetryError, KubernetesError))

        if no_wait:
            client.stdout(
                "Deployed {0} in namespace `{1}`. "
                "Please use `azdata arc postgres server show -n {0}` to check its status."
                    .format(cr.name, cr.namespace))
        else:
            state = None
            if not is_windows():
                with AutomaticSpinner('Deploying {0} in namespace `{1}`'.format(cr.name, cr.namespace),
                                      show_time=True):
                    while state != 'ready' or state is None:
                        time.sleep(5)
                        response = util.retry(lambda: client.apis.kubernetes.get_namespaced_custom_object(
                                                cr.name,
                                                cr.namespace,
                                                group=API_GROUP,
                                                version=API_VERSION,
                                                plural=resource_kind_plural),
                                              retry_count=CONNECTION_RETRY_ATTEMPTS,
                                              retry_delay=RETRY_INTERVAL, retry_method="get namespaced custom object",
                                              retry_on_exceptions=(NewConnectionError, MaxRetryError, KubernetesError))

                        deployed_cr = PostgresqlCustomResource(response)
                        state = deployed_cr.state
                        if state is not None:
                            state = state.lower()
            else:
                client.stdout('Deploying {0} in namespace `{1}`'.format(name, namespace))
                while state != 'ready' or state is None:
                    time.sleep(5)
                    response = util.retry(lambda: client.apis.kubernetes.get_namespaced_custom_object(
                                            cr.name,
                                            cr.namespace,
                                            group=API_GROUP,
                                            version=API_VERSION,
                                            plural=resource_kind_plural),
                                          retry_count=CONNECTION_RETRY_ATTEMPTS,
                                          retry_delay=RETRY_INTERVAL, retry_method="get namespaced custom object",
                                          retry_on_exceptions=(NewConnectionError, MaxRetryError, KubernetesError))

                    deployed_cr = PostgresqlCustomResource(response)
                    state = deployed_cr.state
                    if state is not None:
                        state = state.lower()

            client.stdout('{0} is Ready'.format(cr.name))

    except KubernetesError as e:
        raise CliError(e.message)
    except Exception as e:
        raise CliError(e)


def arc_postgres_server_edit(
        client,
        name,
        # namespace=None,
        path=None,
        workers=None,
        engine_version=None,
        cores_limit=None,
        cores_request=None,
        memory_limit=None,
        memory_request=None,
        extensions=None,
        dev=None,
        port=None,
        no_wait=False,
        engine_settings=None,
        replace_engine_settings=None,
        admin_password=False):
    """
    Edit the configuration of a PostgreSQL server group.
    :param client:
    :param name: The name of the PostgreSQL server group you would like to edit.
    :param path: The path to the source json file for the PostgreSQL server group. This is optional.
    :param namespace: Namespace where the PostgreSQL server group is deployed.
    :param workers: The number of worker nodes to provision in a sharded cluster, or zero for single-node Postgres.
    :param cores_limit: The limit of cores of the managed instance in integer number of vcores.
    :param cores_request: The request for cores of the managed instance in integer number of vcores.
    :param memory_limit: The limit of the capacity of the managed instance in integer amount of memory in GBs.
    :param memory_request: The request for the capacity of the managed instance in integer amount of memory in GBs.
    :param extensions: A comma-separated list of Postgres extensions that should be enabled.
    :param dev: If this is specified, then it is considered a dev instance and will not be billed for.
    :param port: Optional parameter for the service port.
    :param no_wait: If given, the command won't wait for the deployment to be ready before returning.
    :param engine_settings: If given, sets the engine properties
    :param replace_engine_settings: If given, replaces all existing engine settings
    :param admin_password: The admin password for Postgres.
    :return:
    """
    try:
        util.check_and_set_kubectl_context()

        # TODO: Support user supplied namespace when the backend supports it
        namespace = client.profile.active_context.namespace

        # Get the Postgres resource
        old_cr, crd = _get_postgres_custom_object(client, name, namespace, engine_version)
        if old_cr is None:
            raise CliError("PostgreSQL server group {} not found.".format(name))

        if path:
            # Load the spec from a file if provided
            cr = PostgresqlCustomResource(FileUtil.read_json(path))
        else:
            # Otherwise clone the existing resource so we can validate against it
            cr = PostgresqlCustomResource(copy.deepcopy(old_cr.body))

        # Name required
        cr.name = name

        cr.namespace = namespace

        # add scheduling settings
        if cores_limit  is not None:
            cr.default_cores_limit = cores_limit

        if cores_request is not None:
            cr.default_cores_request = cores_request

        if memory_limit  is not None:
            cr.default_memory_limit = memory_limit

        if memory_request  is not None:
            cr.default_memory_request = memory_request

        # Validate cores/memory after both have been set to avoid behavior related to the order of property setting
        #
        cr.validate_memory()
        cr.validate_cores()

        if workers is not None:
            cr.shards = workers

        if extensions:
            extensions = [e.strip() for e in extensions.split(',')]
        else:
            extensions = cr.extensions if extensions is None else []
        if cr.shards and 'citus' not in extensions: # automatically add the 'citus' plugin if sharding is enabled
            extensions.insert(0, 'citus')
        cr.extensions = extensions

        if port:
            cr.service_port = port

        if dev:
            cr.dev = dev

        if engine_settings is not None:
            engine_settings_tuple = (engine_settings, replace_engine_settings)
            cr.default_engine_settings = engine_settings_tuple

        # TODO: Validations on the spec should happen on the backend.
        # Until we have webhook validation configured, we'll do them here.
        # https://sqlhelsinki.visualstudio.com/aris/_workitems/edit/16349
        if cr.shards < old_cr.shards:
            raise CliError("The number of workers cannot be decreased.")

        # Update the admin password if requested
        if admin_password:
            pw = os.environ.get(AZDATA_PASSWORD)
            if not pw:
                if sys.stdin.isatty():
                    pw = prompt_pass('Postgres Server password:', confirm=True, allow_empty=False)
                else:
                    raise ValueError('Please provide a Postgres Server password '
                                     'through the AZDATA_PASSWORD environment variable.')

            client.stdout('Updating password')
            client.apis.kubernetes.patch_secret(namespace, name+'-login-secret', {'stringData': {'password': pw}})

        # Replace CR
        client.apis.kubernetes.replace_namespaced_custom_object(crd, cr)

        if no_wait:
            client.stdout(
                "Updated {0} in namespace `{1}`. "
                "Please use `azdata arc postgres server show -n {0} to check its status."
                    .format(cr.name, cr.namespace))
        else:
            state = None
            if not is_windows():
                with AutomaticSpinner('Updating {0} in namespace `{1}`'.format(cr.name, cr.namespace),
                                      show_time=True):
                    while state != 'ready' or state is None:
                        time.sleep(5)
                        response = client.apis.kubernetes.get_namespaced_custom_object(
                            cr.name, cr.namespace, group=API_GROUP, version=API_VERSION, plural=crd.plural
                        )
                        deployed_cr = PostgresqlCustomResource(response)
                        state = deployed_cr.state
                        if state is not None:
                            state = state.lower()
            else:
                client.stdout('Updating {0} in namespace `{1}`'.format(name, namespace))
                while state != 'ready' or state is None:
                    time.sleep(5)
                    response = client.apis.kubernetes.get_namespaced_custom_object(
                        cr.name, cr.namespace, group=API_GROUP, version=API_VERSION, plural=crd.plural
                    )
                    deployed_cr = PostgresqlCustomResource(response)
                    state = deployed_cr.state
                    if state is not None:
                        state = state.lower()

            client.stdout('{0} is Ready'.format(cr.name))

    except KubernetesError as e:
        raise CliError(e.message)
    except Exception as e:
        raise CliError(e)


def arc_postgres_server_delete(client, name, engine_version=None):
    """
    Delete a PostgreSQL server group.
    :param client:
    :param name: Name of the PostgreSQL server group..
    :return:
    """
    try:
        util.check_and_set_kubectl_context()

        # TODO: Support user supplied namespace when the backend supports it
        namespace = client.profile.active_context.namespace

        cr, crd = _get_postgres_custom_object(client, name, namespace, engine_version)
        if cr is None:
            raise CliError("PostgreSQL server group {} not found.".format(name))

        client.apis.kubernetes.delete_namespaced_custom_object(crd, name, namespace)

        client.stdout("Deleted {} from namespace {}".format(name, namespace))

        client.stdout('Note: Deleting a server group does not remove its associated storage. Reach to your Kubernetes administrator or '+
            'read documentation article "Delete an Azure Arc enabled PostgreSQL Hyperscale server group" for possible next steps.')

    except KubernetesError as e:
        raise CliError(e.message)
    except Exception as e:
        raise CliError(e)


def arc_postgres_server_show(client, name, engine_version=None, path=None):
    """
    Show the details of a PostgreSQL server group.
    :param client:
    :param name: Name of the PostgreSQL server group.
    :param path: A path to a json file where the full specification for the PostgreSQL server group should be written.
    :return:
    """
    try:
        util.check_and_set_kubectl_context()

        # TODO: Support user supplied namespace when the backend supports it
        namespace = client.profile.active_context.namespace

        (cr, _) = _get_postgres_custom_object(client, name, namespace, engine_version)
        if cr is None:
            raise CliError("PostgreSQL server group {} not found.".format(name))
        
        if path:
            if not os.path.isdir(path):
                os.makedirs(path)
            path = os.path.join(path, "{}.json".format(name))
            with open(path, "w") as outfile:
                json.dump(cr.body, outfile, indent=4)
            client.stdout("{0} specification written to {1}".format(name, path))
        else:
            return cr.body

    except KubernetesError as e:
        raise CliError(e.message)
    except Exception as e:
        raise CliError(e)


def arc_postgres_server_list(client):
    """
    List PostgreSQL server groups.
    :param client:
    :return:
    """
    try:
        util.check_and_set_kubectl_context()

        crds = _get_postgres_crds()

        # TODO: Support user supplied namespace when the backend supports it
        namespace = client.profile.active_context.namespace

        items = []
        for crd in crds:
            response = client.apis.kubernetes.list_namespaced_custom_object(
                namespace, group=API_GROUP, version=API_VERSION, plural=crd.spec.names.plural)
            # Temporary, need to discuss with PMs what standardized output we"d like for all partners
            items += response.get("items")

        result = []
        items.sort(key=lambda i:i["kind"] + "\n" + i["metadata"]["name"])
        for item in items:
            cr = PostgresqlCustomResource(item)
            result.append(
                {"name": cr.name,
                 "workers": cr.shards,  # defaults to 0
                 "state": cr.state})

        return result

    except KubernetesError as e:
        raise CliError(e.message)
    except Exception as e:
        raise CliError(e)


def arc_postgres_server_endpoint_list(client, name, engine_version=None):
    """
    List PostgreSQL server groups.
    :param client:
    :param name: Name of the PostgreSQL server group.
    :return:
    """
    try:
        util.check_and_set_kubectl_context()

        # TODO: Support user supplied namespace when the backend supports it
        namespace = client.profile.active_context.namespace

        (cr, _) = _get_postgres_custom_object(client, name, namespace, engine_version)
        if cr is None:
            raise CliError("PostgreSQL server group {} not found.".format(name))

        # Temporary solution until server side logic exists:
        # In the future, this logic should not be part of the client.
        # The custom resource, should list any needed endpoints about it
        # as part of its status. Otherwise, there will be no means of
        # retrieving them for other kubernetes clients, such as kubectl.
        endpoints = []

        if cr.status:
            descrip_str = "Description"
            endpoint_str = "Endpoint"

            # Connection string
            ext_endpoint = cr.status.get("externalEndpoint")
            if ext_endpoint:
                connection_str = "postgresql://postgres:<replace with password>@{}".format(ext_endpoint)
            else:
                connection_str = "Not yet available"
            endpoints.append({descrip_str: "PostgreSQL Instance", endpoint_str: connection_str})

            # Retrieve management proxy service
            service = client.apis.kubernetes.get_service(namespace, MGMT_PROXY)
            proxy_endpoint = client.apis.kubernetes.get_service_endpoint(namespace, service, app="mgmtproxy")

            # Logs
            logs_str = "{0}/kibana/app/kibana#/discover?_a=" \
                       "(query:(language:kuery,query:'custom_resource_name:\"{1}\"'))".format(proxy_endpoint, name)
            endpoints.append({descrip_str: "Log Search Dashboard", endpoint_str: logs_str})

            # Metrics
            metrics_str = "{0}/grafana/d/postgres-metrics?var-Namespace={1}&var-Name={2}" \
                .format(proxy_endpoint, namespace, name)
            endpoints.append({descrip_str: "Metrics Dashboard", endpoint_str: metrics_str})

        return endpoints

    except KubernetesError as e:
        raise CliError(e.message)
    except Exception as e:
        raise CliError(e)


def arc_postgres_server_config_init(client, path, engine_version=None):
    """
    Returns a package of crd.json and spec-template.json.
    :param client:
    :param path:
    :return:
    """
    try:
        if not os.path.isdir(path):
            os.makedirs(path, exist_ok=True)
        
        resource_kind = _get_resource_kind_by_engine_version(engine_version)

        if not is_windows():
            with AutomaticSpinner("Fetching {0} template".format(resource_kind),
                                  show_time=True):
                crd_response = client.apis.controller.resource_crd_get(resource_kind)
                spec_response = client.apis.controller.resource_spec_get(resource_kind)
        else:
            crd_response = client.apis.controller.resource_crd_get(resource_kind)
            spec_response = client.apis.controller.resource_spec_get(resource_kind)

        crd_pretty = json.dumps(crd_response, indent=4)
        spec_pretty = json.dumps(spec_response, indent=4)

        with open(os.path.join(path, "crd.json"), "w") as output:
            output.write(crd_pretty)

        with open(os.path.join(path, "spec.json"), "w") as output:
            output.write(spec_pretty)

        client.stdout("{0} template created in directory: {1}".format(resource_kind, path))

    except Exception as e:
        raise CliError(e)


def arc_postgres_server_config_add(client, path, json_values):
    """
    Add new key and value to the given config file
    :param client:
    :param path:
    :param json_values:
    :return:
    """
    try:
        config_object = DeploymentConfigUtil.config_add(path, json_values)
        DeploymentConfigUtil.write_config_file(path, config_object)
    except Exception as e:
        raise CliError(e)


def arc_postgres_server_config_replace(client, path, json_values):
    """
    Replace the value of a given key in the given config file
    :param client:
    :param path:
    :param json_values:
    :return:
    """
    try:
        config_object = DeploymentConfigUtil.config_replace(path, json_values)
        DeploymentConfigUtil.write_config_file(path, config_object)
    except Exception as e:
        raise CliError(e)


def arc_postgres_server_config_remove(client, path, json_path):
    """
    Remove a key from the given config file
    :param client:
    :param path:
    :param json_path:
    :return:
    """
    try:
        config_object = DeploymentConfigUtil.config_remove(path, json_path)
        DeploymentConfigUtil.write_config_file(path, config_object)
    except Exception as e:
        raise CliError(e)


def arc_postgres_server_config_patch(client, path, patch_file):
    """
    Patch a given file against the given config file
    :param client:
    :param path:
    :param patch_file:
    :return:
    """
    try:
        config_object = DeploymentConfigUtil.config_patch(path, patch_file)
        DeploymentConfigUtil.write_config_file(path, config_object)
    except Exception as e:
        raise CliError(e)

def arc_postgres_backup_create(client, server_name, backup_name=None, no_wait=False):
    """
    Create a backup of a PostgreSQL server group.
    :param client:
    :param server_name: Name of the PostgreSQL server group.
    :param backup_name: Name of the backup. This parameter is optional.
    :return:
    """
    try:
        util.check_and_set_kubectl_context()

        # TODO: Support user supplied namespace when the backend supports it
        namespace = client.profile.active_context.namespace
        (cr, _) = _get_postgres_custom_object(client, server_name, namespace, None)
        if cr == None:
            raise CliError("PostgreSQL Server group {0} not found.".format(server_name))

        server_group_id = cr.metadata["uid"]

        backup = client.apis.controller.postgres_server_backup_create(namespace=namespace, server_group_id=server_group_id, backup_name=backup_name)

        delay = 5
        while not no_wait:
            time.sleep(delay)
            backup = client.apis.controller.postgres_server_backup_show(namespace=namespace, server_group_id=server_group_id, backup_id=backup['receipt']['id'])
            if backup['progress'] not in ("Pending", "Active"): break
            delay *= 2
            if delay > 60: delay = 60

        return _format_backup(backup)
    except Exception as e:
        raise CliError(e)


def arc_postgres_backup_restore(client, server_name, backup_id):
    """
    Restore a backup of a PostgreSQL server group.
    :param client:
    :param server_name: Name of the PostgreSQL server group.
    :param backup_id: ID of the backup.
    :return:
    """
    try:
        util.check_and_set_kubectl_context()

        # TODO: Support user supplied namespace when the backend supports it
        namespace = client.profile.active_context.namespace
        (cr, _) = _get_postgres_custom_object(client, server_name, namespace, None)
        if cr == None:
            raise CliError("PostgreSQL Server group {0} not found.".format(server_name))

        server_group_id = cr.metadata["uid"]

        status = client.apis.controller.postgres_server_backup_restore(namespace=namespace, server_group_id=server_group_id, backup_id=backup_id)

        start_time, delay = time.time(), 5
        while True: # wait for the restore to complete
            time.sleep(delay)
            status = client.apis.controller.postgres_server_backup_restore_status(namespace=namespace, server_group_id=server_group_id)
            if status['progress'] not in ("Pending", "Active"): break
            delay *= 2
            if delay > 60: delay = 60

        return _format_restore(status)
    except Exception as e:
        raise CliError(e)


def arc_postgres_backup_list(client, server_name):
    """
    Show all backups for a PostgreSQL server group.
    :param client:
    :param server_name: The name of the server group.
    :return:
    """
    try:
        util.check_and_set_kubectl_context()

        # TODO: Support user supplied namespace when the backend supports it
        namespace = client.profile.active_context.namespace

        (cr, _) = _get_postgres_custom_object(client, server_name, namespace, None)
        if cr == None:
            raise CliError("PostgreSQL Server group {0} not found. No backup was retrieved.".format(server_name))

        backups = client.apis.controller.postgres_server_backup_list(namespace=namespace, server_group_id=cr.metadata["uid"])
        return [_format_backup(b) for b in backups]
    except Exception as e:
        raise CliError(e)

def arc_postgres_backup_delete(client, server_name, backup_name=None, backup_id=None):
    """
    Delete the backup for a PostgreSQL server group.
    :param client:
    :param server_name: The name of the server group.
    :param backup_name: The name of the backup to be deleted.
    :param backup_id: The id of the backup to be deleted.
    :return:
    """
    try:
        util.check_and_set_kubectl_context()

        # TODO: Support user supplied namespace when the backend supports it
        namespace = client.profile.active_context.namespace

        (cr, _) = _get_postgres_custom_object(client, server_name, namespace, None)
        if cr == None:
            raise CliError("PostgreSQL Server group {0} not found.".format(server_name))

        if backup_name is not None:
            backup_id = _get_backup_id_by_name(client, namespace, server_name, backup_name)

        backup = client.apis.controller.postgres_server_backup_delete(namespace=namespace, server_group_id=cr.metadata["uid"], backup_id=backup_id)
        return _format_backup(backup)
    except Exception as e:
        raise CliError(e)

def _get_backup_id_by_name(client, namespace, server_name, backup_name) -> str:
    backups_list = arc_postgres_backup_list(client, server_name)
    backups = [b for b in backups_list if b['name'] == backup_name]
    if len(backups) == 0:
        raise ValueError("Could not find a backup with name '%s' for server %s." % (backup_name, server_name))
    elif len(backups) > 1:
        raise ValueError("There are multiple backups named '%s'. Use -id parameter instead." % backup_name)

    return backups[0]['ID']

def _format_backup(backup):
    result = OrderedDict()
    result['name'] = backup.get('backup', {}).get('label', '')
    result['ID'] = backup['receipt']['id']
    result['state'] = backup['progress']
    result['timestamp'] = backup['timestamp']
    if backup['status'] is not None and not backup['status']['result']:
        result['error'] = backup['status']['message']
    return result

def _format_restore(restore):
    result = OrderedDict()
    result['ID'] = restore['receipt']['id']
    result['state'] = restore['progress']
    if restore['status'] is not None and not restore['status']['result']:
        result['error'] = restore['status']['message']
    return result

# def arc_postgres_server_restart(client, name, namespace=None):
#     """
#     Restart a PostgreSQL server group.
#     :param client:
#     :param name: Name of the PostgreSQL server group.
#     :param namespace: Namespace where the PostgreSQL server groups are deployed.
#     :return:
#     """
#     try:
#         client.stdout(COMMAND_UNIMPLEMENTED)
#     except Exception as e:
#         raise CliError(e)


# ------------------------------------------------------------------------------
# Database Commands - Unimplemented
# ------------------------------------------------------------------------------

# def arc_postgres_database_create(client, namespace, name, server_group_name, sharded=False, owner=None):
#     """
#     Create a database within a PostgreSQL server group.
#     :param client:
#     :param namespace: Namespace where the PostgreSQL server group is deployed.
#     :param name: Name of the database.
#     :param server_group_name: Name of the PostgreSQL server group.
#     :param sharded: An optional boolean that indicates whether the database should support sharding.
#     :param owner: An optional name of a user that should own the database.
#     :return:
#     """
#     try:
#         client.stdout(COMMAND_UNIMPLEMENTED)
#     except Exception as e:
#         raise CliError(e)


# def arc_postgres_database_delete(client, namespace, name, server_group_name):
#     """
#     Delete a database from a PostgreSQL server group.
#     :param client:
#     :param namespace: Namespace where the PostgreSQL server group is deployed.
#     :param name: Name of the database.
#     :param server_group_name: Name of the PostgreSQL server group.
#     :return:
#     """
#     try:
#         client.stdout(COMMAND_UNIMPLEMENTED)
#     except Exception as e:
#         raise CliError(e)
#
#
# def arc_postgres_database_show(client, namespace, name, server_group_name):
#     """
#     Show the details of a PostgreSQL database.
#     :param client:
#     :param namespace: Namespace where the PostgreSQL server group is deployed.
#     :param name: Name of the database.
#     :param server_group_name: Name of the PostgreSQL server group.
#     :return:
#     """
#     try:
#         client.stdout(COMMAND_UNIMPLEMENTED)
#     except Exception as e:
#         raise CliError(e)
#
#
# def arc_postgres_database_edit(client, namespace, name, server_group_name, sharded=False, owner=None):
#     """
#     Edit the configuration of a database within a PostgreSQL server group.
#     :param client:
#     :param namespace: Namespace where the PostgreSQL server group is deployed.
#     :param name: Name of the database.
#     :param server_group_name: Name of the PostgreSQL server group.
#     :param sharded: An optional boolean that indicates whether the database should support sharding.
#     :param owner: An optional name of a user that should own the database.
#     :return:
#     """
#     try:
#         client.stdout(COMMAND_UNIMPLEMENTED)
#     except Exception as e:
#         raise CliError(e)
#
#
# def arc_postgres_database_list(client, namespace, server_group_name):
#     """
#     List databases within a PostgreSQL server group.
#     :param client:
#     :param namespace: Namespace where the PostgreSQL server group is deployed.
#     :param server_group_name: Name of the PostgreSQL server group.
#     :return:
#     """
#     try:
#         client.stdout(COMMAND_UNIMPLEMENTED)
#     except Exception as e:
#         raise CliError(e)
#
#
#
# # ------------------------------------------------------------------------------
# # Role Commands - Unimplemented
# # ------------------------------------------------------------------------------
#
# def arc_postgres_role_create(client, namespace, name, server_group):
#     """
#     Create a role within a PostgreSQL server group.
#     :param client:
#     :param namespace: Namespace where the PostgreSQL server group is deployed.
#     :param name: Name of the role.
#     :param server_group: Name of the PostgreSQL server group.
#     :return:
#     """
#     try:
#         client.stdout(COMMAND_UNIMPLEMENTED)
#     except Exception as e:
#         raise CliError(e)
#
#
# def arc_postgres_role_delete(client, namespace, name, server_group):
#     """
#     Delete a role from a PostgreSQL server group.
#     :param client:
#     :param namespace: Namespace where the PostgreSQL server group is deployed.
#     :param name: Name of the role.
#     :param server_group: Name of the PostgreSQL server group.
#     :return:
#     """
#     try:
#         client.stdout(COMMAND_UNIMPLEMENTED)
#     except Exception as e:
#         raise CliError(e)
#
#
# def arc_postgres_role_list(client, namespace, server_group):
#     """
#     List roles within a PostgreSQL server group.
#     :param client:
#     :param namespace: Namespace where the PostgreSQL server group is deployed.
#     :param server_group: Name of the PostgreSQL server group.
#     :return:
#     """
#     try:
#         client.stdout(COMMAND_UNIMPLEMENTED)
#     except Exception as e:
#         raise CliError(e)
#
#
# def arc_postgres_role_show(client, namespace, name, server_group):
#     """
#     Show the details of a PostgreSQL role.
#     :param client:
#     :param namespace: Namespace where the PostgreSQL server group is deployed.
#     :param name: Name of the role.
#     :param server_group: Name of the PostgreSQL server group.
#     :return:
#     """
#     try:
#         client.stdout(COMMAND_UNIMPLEMENTED)
#     except Exception as e:
#         raise CliError(e)
#
#
# # ------------------------------------------------------------------------------
# # User Commands - Unimplemented
# # ------------------------------------------------------------------------------
#
# def arc_postgres_user_create(client, namespace, name, password, roles, server_group):
#     """
#     Create a user within a PostgreSQL server group.
#     :param client:
#     :param namespace: Namespace where the PostgreSQL server group is deployed.
#     :param name: Name of the user.
#     :param password: An optional password for the user. If omitted, a password will be generated.
#     :param roles: An optional comma-separated list of roles that the user should belong to.
#     :param server_group: Name of the PostgreSQL server group.
#     :return:
#     """
#     try:
#         client.stdout(COMMAND_UNIMPLEMENTED)
#     except Exception as e:
#         raise CliError(e)
#
#
# def arc_postgres_user_edit(client, namespace, name, password, roles, server_group):
#     """
#     Edit the configuration of a user within a PostgreSQL server group.
#     :param client:
#     :param namespace: Namespace where the PostgreSQL server group is deployed.
#     :param name: Name of the user.
#     :param password: An optional password for the user. If omitted, a password will be generated.
#     :param roles: An optional comma-separated list of roles that the user should belong to.
#     :param server_group: Name of the PostgreSQL server group.
#     :return:
#     """
#     try:
#         client.stdout(COMMAND_UNIMPLEMENTED)
#     except Exception as e:
#         raise CliError(e)
#
#
# def arc_postgres_user_delete(client, namespace, name, server_group):
#     """
#     Delete a user from a PostgreSQL server group.
#     :param client:
#     :param namespace: Namespace where the PostgreSQL server group is deployed.
#     :param name: Name of the user.
#     :param server_group: Name of the PostgreSQL server group.
#     :return:
#     """
#     try:
#         client.stdout(COMMAND_UNIMPLEMENTED)
#     except Exception as e:
#         raise CliError(e)
#
#
# def arc_postgres_user_list(client, namespace, server_group):
#     """
#     List users within a PostgreSQL server group.
#     :param client:
#     :param namespace: Namespace where the PostgreSQL server group is deployed.
#     :param server_group: Name of the PostgreSQL server group.
#     :return:
#     """
#     try:
#         client.stdout(COMMAND_UNIMPLEMENTED)
#     except Exception as e:
#         raise CliError(e)
#
#
# def arc_postgres_user_show(client, namespace, name, server_group):
#     """
#     Show the details of a PostgreSQL user.
#     :param client:
#     :param namespace: Namespace where the PostgreSQL server group is deployed.
#     :param name: Name of the user.
#     :param server_group: Name of the PostgreSQL server group.
#     :return:
#     """
#     try:
#         client.stdout(COMMAND_UNIMPLEMENTED)
#     except Exception as e:
#         raise CliError(e)
#
#
# # ------------------------------------------------------------------------------
# # Volume Commands - Unimplemented
# # ------------------------------------------------------------------------------
#
# def arc_postgres_volume_delete(client, namespace, server_group=None):
#     """
#     Delete all inactive volume claims associated with one or all deleted PostgreSQL server groups.
#     :param client:
#     :param namespace: Namespace where the PostgreSQL server group is deployed.
#     :param server_group: Name of the PostgreSQL server group.
#     :return:
#     """
#     try:
#         client.stdout(COMMAND_UNIMPLEMENTED)
#     except Exception as e:
#         raise CliError(e)
#
#
# def arc_postgres_volume_list(client, namespace, server_group):
#     """
#     List volume claims of one or all PostgreSQL server groups.
#     :param client:
#     :param namespace: Namespace where the PostgreSQL server group is deployed.
#     :param server_group: Name of the PostgreSQL server group.
#     :return:
#     """
#     try:
#         client.stdout(COMMAND_UNIMPLEMENTED)
#     except Exception as e:
#         raise CliError(e)
#
#
# def arc_postgres_volume_show(client, namespace, server_group):
#     """
#     Show the details of a PostgreSQL volume.
#     :param client:
#     :param namespace: Namespace where the PostgreSQL server group is deployed.
#     :param server_group: Name of the PostgreSQL server group.
#     :return:
#     """
#     try:
#         client.stdout(COMMAND_UNIMPLEMENTED)
#     except Exception as e:
#         raise CliError(e)


def _get_postgres_crds():
    """
    Returns a list of CRDs for the supported postgresql engine versions.
    :return:
    """
    api = k8sClient.ApiextensionsV1beta1Api()
    crds = api.list_custom_resource_definition()
    return [crd for crd in crds.items 
        if crd.spec.names.categories is not None and RESOURCE_CATEGORY in crd.spec.names.categories]

def _get_postgres_custom_object(client, name, namespace, engine_version: int):
    """
    Returns the custom object and the corresponding CRD as a tuple for the PostgreSQL server group identified by name and engine version.
    It's possible to create multipe PostgreSQL server groups with the same name but different engine version.
    Name and engine version can uniquely identify a PostgreSQL server group in a namespace.
    :param client:
    :param name: The name of the instance.
    :param namespace: Namespace where the PostgreSQL server group is deployed.
    :param engine_version: The postgresql engine version.
    :return: None will be returned if the PostgreSQL server group cannot be found.
             CliError will be raised if there are multiple PostgreSQL server groups found.
    """
    if engine_version is None:
        kinds = [c.spec.names.kind for c in _get_postgres_crds()]
    else:
        kinds = [_get_resource_kind_by_engine_version(engine_version)]
    cr, crd = None, None
    for k in kinds:
        try:
            c = CustomResourceDefinition(client.apis.controller.resource_crd_get(k))
            result = client.apis.kubernetes.get_namespaced_custom_object(name, namespace, crd=c)
        except K8sApiException as e:
            if e.status == http_status_codes.not_found:
                continue
        if cr is not None:
            raise CliError("Multiple PostgreSQL server groups found. " + 
                "They have the same name but different engine version. " + 
                "Please specify --engine-version.")
        cr = PostgresqlCustomResource(result)
        crd = c
    return cr, crd

def _get_postgres_custom_objects(client, name, namespace):
    """
    Returns a list of custom objects for the PostgreSQL server groups by name.
    It's possible to create multipe PostgreSQL server groups with the same name but different engine version.
    :param client:
    :param name: The name of the instance.
    :param namespace: Namespace where the PostgreSQL server group is deployed.
    :return:
    """
    kinds = [c.spec.names.kind for c in _get_postgres_crds()]
    crs = []
    for k in kinds:
        try:
            crd = CustomResourceDefinition(client.apis.controller.resource_crd_get(k))
            result = client.apis.kubernetes.get_namespaced_custom_object(name, namespace, crd=crd)
        except K8sApiException as e:
            if e.status == http_status_codes.not_found:
                continue
        crs.append(PostgresqlCustomResource(result))
    return crs

def _get_resource_kind_by_engine_version(engine_version: int):
    """
    Returns the resource kind defined in the corresponding CRD for the specified postgresql engine version.
    It's possible to create multipe PostgreSQL server groups with the same name but different engine version.
    :param engine_version: The postgresql engine version. If not specified, the latest supported engine version will be used.
    :return:
    """
    if engine_version == 11:
        return RESOURCE_KIND_11
    elif engine_version == 12 or engine_version is None: # Default to the latest supported version
        return RESOURCE_KIND_12
    else:
        raise CliError("Unsupported engine version {}".format(engine_version))
