# ------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ------------------------------------------------------------------------------

import re

from azdata.cli.core.models.custom_resource import CustomResource
from azdata.cli.commands.postgres.constants import (
    POSTGRES_MIN_CORES_SIZE,
    POSTGRES_MIN_MEMORY_SIZE)
from azdata.cli.core.models.kube_resource import (
    CpuResource,
    MemoryResource)


class PostgresqlCustomResource(CustomResource):
    """
    Internal Postgres Custom Resource object to be used for deployments.
    """
    def __init__(self, cr_object):
        """
        Initializes a CR object with the given json.
        :param cr_object: The custom resource json.
        """
        super().__init__(cr_object)

    @property
    def default_memory_request(self):
        """
        Returns the memory request property of postgres if exists else None if intermediate params does not exist
        :return:
        """
        return self.spec.get("scheduling", {}).get("default", {}).get("resources", {}).get("requests", {}).get("memory")

    @default_memory_request.setter
    def default_memory_request(self, request):
        """
        Sets the memory request property of postgres.
        :param self:
        :param request:
        :return:
        """
        if request == '':
            memory_request = self.default_memory_request
            if memory_request:
                self.spec.get("scheduling").get("default").get("resources").get("requests").pop("memory")
        else:
            request_size = MemoryResource(request)
            if request_size < POSTGRES_MIN_MEMORY_SIZE:
                raise ValueError("Memory request must be at least %s" % POSTGRES_MIN_MEMORY_SIZE.quantity)

            self.spec.setdefault("scheduling", {}).setdefault("default", {}).setdefault("resources", {}).setdefault("requests", {})["memory"] = request

    @property
    def default_cores_request(self):
        """
        Returns the vcore request property for postgres if exists else None if intermediate params does not exist
        :return:
        """
        return self.spec.get("scheduling", {}).get("default", {}).get("resources", {}).get("requests", {}).get("cpu")

    @default_cores_request.setter
    def default_cores_request(self, request):
        """
        Sets the vcore request property for postgres.
        :param self:
        :param request:
        :return:
        """
        if request == '':
            cores_request = self.default_cores_request
            if cores_request:
                self.spec.get("scheduling").get("default").get("resources").get("requests").pop("cpu")
        else:
            request_size = CpuResource(request)
            if request_size < POSTGRES_MIN_CORES_SIZE:
                raise ValueError("Cores request cannot be less than %s" % POSTGRES_MIN_CORES_SIZE.quantity)

            self.spec.setdefault("scheduling", {}).setdefault("default", {}).setdefault("resources", {}).setdefault("requests", {})["cpu"] = request

    @property
    def default_cores_limit(self):
        """
        Returns the vcore limit property for postgres if exists else None if intermediate params does not exist
        :return:
        """
        return self.spec.get("scheduling", {}).get("default", {}).get("resources", {}).get("limits", {}).get("cpu")

    @default_cores_limit.setter
    def default_cores_limit(self, lim):
        """
        Sets the vcore limit property for postgres.
        :param self:
        :param lim:
        :return:
        """
        if lim == '':
            cores_limit = self.default_cores_limit
            if cores_limit:
                self.spec.get("scheduling", {}).get("default", {}).get("resources", {}).get("limits", {}).pop("cpu")
        else:
            limit_size = CpuResource(lim)
            if limit_size < POSTGRES_MIN_CORES_SIZE:
                raise ValueError("Cores limit must be at least %s" % POSTGRES_MIN_CORES_SIZE.quantity)

            self.spec.setdefault("scheduling", {}).setdefault("default", {}).setdefault("resources", {}).setdefault("limits", {})["cpu"] = lim

    @property
    def default_memory_limit(self):
        """
        Returns the memory limit property if exists else None if intermediate params does not exist
        :return:
        """
        return self.spec.get("scheduling", {}).get("default", {}).get("resources", {}).get("limits", {}).get("memory")

    @default_memory_limit.setter
    def default_memory_limit(self, lim):
        """
        Sets the memory limit property.
        :param self:
        :param lim:
        :return:
        """
        if lim == '':
            memory_limit = self.default_memory_limit
            if memory_limit:
                self.spec.get("scheduling", {}).get("default", {}).get("resources", {}).get("limits", {}).pop("memory")
        else:
            limit_size = MemoryResource(lim)
            if limit_size < POSTGRES_MIN_MEMORY_SIZE:
                raise ValueError("Memory limit must be at least %s" % POSTGRES_MIN_MEMORY_SIZE.quantity)

            self.spec.setdefault("scheduling", {}).setdefault("default", {}).setdefault("resources", {}).setdefault("limits", {})["memory"] = lim

    @property
    def default_engine_settings(self):
        """
        Returns the default engine settings if exists else returns empty dictionary.
        :return:
        """
        return self.spec.get("engine", {}).get("settings", {}).get("default", {})

    @default_engine_settings.setter
    def default_engine_settings(self, engine_settings_tuple):
        """
        Sets the engine settings property.
        :param self:
        :param engine_settings_tuple: tuple containing engine_settings and replace_engine_settings
        :return:
        """
        self.spec.setdefault("engine", {}).setdefault('settings', {})['default'] = self._get_new_engine_settings(engine_settings_tuple[0], engine_settings_tuple[1])

    def _get_new_engine_settings(self, engine_settings, replace_engine_settings : bool):
        original_settings = dict(self.default_engine_settings)

        if replace_engine_settings:
            original_settings = {}

        new_settings = dict()
        for x, y in self._get_engine_settings_list(engine_settings):
            if not y:
                new_settings[x] = None
            elif y[0] != '"':
                new_settings[x] = y
            else:
                new_settings[x] = y[1:-1].replace('""', "")
        for k,v in new_settings.items():
            if v is None:
                original_settings.pop(k, None)
            else:
                original_settings[k] = v

        return original_settings

    @staticmethod
    def _get_engine_settings_list(engine_settings):
        return re.findall(r'\s*([^\s=,]+)\s*=\s*("(?:[^"]|"")*"|[^\s",]*),?', engine_settings)

    @property
    def engine(self):
        """
        Returns the engine property.
        :param self:
        :return:
        """
        return self.spec.get("engine")

    @engine.setter
    def engine(self, e):
        """
        Sets the engine property.
        :param self:
        :param e:
        :return:
        """
        self.spec["engine"] = e

    @property
    def extensions(self):
        """
        Returns the extensions property as a list of extension names.
        :param self:
        :return:
        """
        return [p['name'] for p in (self.engine or {}).get('extensions') or []]

    @extensions.setter
    def extensions(self, extensions):
        """
        Sets the extensions property from a list of extension names.
        :param self:
        :param extensions:
        :return:
        """
        if len(extensions) != 0:
            if self.engine is None:
                self.engine = {}
            self.engine['extensions'] = [{'name':p} for p in extensions]
        elif self.engine:
            self.engine.pop('extensions', None)

    def validate_memory(self):
        """
        @override
        """
        request = self.default_memory_request
        limit = self.default_memory_limit
        if request and limit:
            request_size = MemoryResource(request)
            limit_size = MemoryResource(limit)
            if limit_size < request_size:
                raise ValueError("Memory request of %s cannot exceed memory limit of %s" % (request, limit))

    def validate_cores(self):
        """
        @override
        """
        request = self.default_cores_request
        limit = self.default_cores_limit
        if request and limit:
            cores_limit = CpuResource(limit)
            cores_request = CpuResource(request)
            if cores_limit < cores_request:
                raise ValueError("Cores request of %s cannot exceed cores limit of %s" % (request, limit))
