import re

from azdata.cli.core.models.custom_resource import CustomResource
from azdata.cli.core.models.kube_resource import (
    MemoryResource,
    CpuResource)
from azdata.cli.commands.sqlmi.constants import (
    SQLMI_MIN_MEMORY_SIZE,
    SQLMI_MIN_CORES_SIZE)


class SqlmiCustomResource(CustomResource):
    """
    Internal Sqlmi Custom Resource object to be used for deployments.
    """
    arc_sql_name_max_length = 13
    arc_sql_name_requirements = "Sql instance names must contain only alphanumeric characters or '-', be lowercase, and" \
                                " start/end with an alphanumeric character"

    def __init__(self, cr_object):
        """
        Initializes a CR object with the given json.
        :param cr_object: The custom resource json.
        """
        super().__init__(cr_object)

    @CustomResource.name.setter
    def name(self, n):
        """
        @override
        """
        if not n:
            raise ValueError("Name cannot be empty or None")

        if len(n) > self.arc_sql_name_max_length:
            raise ValueError("Name '{}' exceeds {} character length limit".format(n, self.arc_sql_name_max_length))

        if not n[0].isalnum() or not n[-1].isalnum():
            raise ValueError(self.arc_sql_name_requirements)

        for c in n:
            if not c.isalnum() and c != '-':
                raise ValueError(self.arc_sql_name_requirements)

        CustomResource.name.fset(self, n)

    @CustomResource.memory_request.setter
    def memory_request(self, request):
        """
        @override
        """
        if request == '':
            CustomResource.memory_request.fset(self, request)
            return

        request_size = MemoryResource(request)
        if request_size < SQLMI_MIN_MEMORY_SIZE:
            raise ValueError("Memory request must be at least 2Gi")

        if self.requests:
            self.requests["memory"] = request
        else:
            self.spec["requests"] = {"memory": request}

    @CustomResource.memory_limit.setter
    def memory_limit(self, limit):
        """
        @override
        """
        if limit == '':
            CustomResource.memory_limit.fset(self, limit)
            return

        limit_size = MemoryResource(limit)
        if limit_size < SQLMI_MIN_MEMORY_SIZE:
            raise ValueError("Memory limit must be at least 2Gi")

        if self.limits:
            self.limits["memory"] = limit
        else:
            self.spec["limits"] = {"memory": limit}

    @CustomResource.cores_limit.setter
    def cores_limit(self, limit):
        """
        @override
        """
        if limit == '':
            CustomResource.cores_limit.fset(self, limit)
            return

        limit_size = CpuResource(limit)
        if limit_size < SQLMI_MIN_CORES_SIZE:
            raise ValueError("Cores limit cannot be below %s" % SQLMI_MIN_CORES_SIZE.quantity)

        if self.limits:
            self.limits["vcores"] = limit
        else:
            self.spec["limits"] = {"vcores": limit}

    @CustomResource.cores_request.setter
    def cores_request(self, request):
        """
        @override
        """
        if request == '':
            CustomResource.cores_request.fset(self, request)
            return

        request_size = CpuResource(request)
        if request_size < SQLMI_MIN_CORES_SIZE:
            raise ValueError("Cores request cannot be below %s" % SQLMI_MIN_CORES_SIZE.quantity)

        if self.requests:
            self.requests["vcores"] = request
        else:
            self.spec["requests"] = {"vcores": request}

