import json
import yaml

from kubernetes.client.rest import ApiException as K8sApiException
from azdata.cli.core.logging import get_logger
import azdata.cli.core.kubernetes as kubernetes_util

logger = get_logger(__name__)

def is_controller_upgrading(cluster_name):
    """
    Check the upgrade config map to see if the controller or controller database are being upgraded
    """
    upgrade_config_map = "controller-upgrade-configmap"

    controller_upgrade_states = ["UpgradingController", "UpgradingControllerDb", "WaitingForControllerDb",
                "RollingBackController", "RollingBackControllerDb", "WaitingForControllerDbRollback", "WaitingForStableUpgrade"]

    try:
        config_map = kubernetes_util.get_config_map(cluster_name, upgrade_config_map)

        upgrade_config = json.loads(config_map.data["controller-upgrade"])

        state = upgrade_config["currentState"]

        if state in controller_upgrade_states:
            return True
        else:
            return False
    except K8sApiException as e:
        logger.error(e.body)
        raise

def is_control_plane_upgrading(cluster_name):
    """
    Check the upgrade config map to see if the control plane is being upgraded
    """
    upgrade_config_map = "controller-upgrade-configmap"

    try:
        config_map = kubernetes_util.get_config_map(cluster_name, upgrade_config_map)

        upgrade_config = json.loads(config_map.data["controller-upgrade"])

        state = upgrade_config["currentState"]

        if state == "UpgradingControlPlane":
            return True
        else:
            return False
    except K8sApiException as e:
        logger.error(e.body)
        raise

def check_controller_upgrade_failed(cluster_name):
    """
    Check the upgrade config map to see if the controller upgrade failed
    """
    upgrade_config_map = "controller-upgrade-configmap"

    try:
        config_map = kubernetes_util.get_config_map(cluster_name, upgrade_config_map)

        upgrade_config = json.loads(config_map.data["controller-upgrade"])

        state = upgrade_config["currentState"]
        previousState = upgrade_config["previousState"]

        if state == "NoUpgradeInProgress" and previousState == "RollingBackController":
            return True
        else:
            return False
    except K8sApiException as e:
        logger.error(e.body)
        raise

def update_upgrade_timeouts(cluster_name, controller_timeout, stability_threshold, component_timeout):
    """
    Update the timeouts for upgrades

    The following timeouts can be set:
    controller_timeout:  number of minute allowed for the controller or controller db to upgrade
    stability_threshold: number of minutes required before marking an upgrade as stable
    component_timeout:   number of minutes allowed for each subsequent phase of the upgrade
    """
    upgrade_config_map = "controller-upgrade-configmap"

    try:
        config_map = kubernetes_util.get_config_map(cluster_name, upgrade_config_map)

        upgrade_config = json.loads(config_map.data["controller-upgrade"])

        if stability_threshold:
            upgrade_config["stableUptimeThresholdInMinutes"] = int(stability_threshold)

        if controller_timeout:
            upgrade_config["controllerUpgradeTimeoutInMinutes"] = int(controller_timeout)

        if stability_threshold or controller_timeout:
            controller_timeout = upgrade_config["controllerUpgradeTimeoutInMinutes"]
            stability_threshold = upgrade_config["stableUptimeThresholdInMinutes"]
            
            # The total amount of time we allow for the controller and controllerdb to finish 
            # upgrading is decided by the amount of time we allow for each component as well as
            # the stability threshold
            #
            upgrade_config["totalUpgradeTimeoutInMinutes"] = controller_timeout * 2 + stability_threshold

        if component_timeout:
            upgrade_config["componentUpgradeTimeoutInMinutes"] = int(component_timeout)

        config_map.data["controller-upgrade"] = json.dumps(upgrade_config)

        kubernetes_util.patch_config_map(cluster_name, upgrade_config_map, config_map)

    except ValueError as e:
        raise Exception("Upgrade timeouts must be integers")
    except K8sApiException as e:
        logger.error(e.body)
        raise
